#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" This module contains meta functionality for the ``Tensor`` type. """
from abc import abstractmethod
from typing import List, Optional, Any, Tuple, Dict, Union

import torch
import numpy as np

from asta.utils import shapecheck, attrcheck
from asta.parser import parse_subscript
from asta.classes import SubscriptableMeta, GenericMeta
from asta.constants import (
    EllipsisType,
    TORCH_DIM_TYPES,
    TORCH_DTYPE_MAP,
)

# pylint: disable=unidiomatic-typecheck, too-few-public-methods, too-many-nested-blocks


class _TensorMeta(SubscriptableMeta):
    """ A meta class for the ``Tensor`` class. """

    shape: tuple
    dtype: torch.dtype
    kwattrs: Dict[str, Any]

    @classmethod
    @abstractmethod
    def _after_subscription(cls, item: Any) -> None:
        """ Method signature for subscript argument processing. """
        raise NotImplementedError

    def __getitem__(cls, item: Any) -> GenericMeta:
        """ Defer to the metaclass which calls ``cls._after_subscription()``. """
        return SubscriptableMeta.__getitem__(cls, item)

    def __eq__(cls, other: Any) -> bool:
        """ If the dtypes and shapes match, they should be equal. """
        if not isinstance(other, _TensorMeta):
            return False
        if cls.shape != other.shape or cls.dtype != other.dtype:
            return False
        return True

    def __instancecheck__(cls, inst: Any) -> bool:
        """ Support expected behavior for ``isinstance(<tensor>, Tensor[<args>])``. """
        assert hasattr(cls, "shape")
        assert hasattr(cls, "dtype")
        assert hasattr(cls, "kwattrs")
        match = False
        if isinstance(inst, torch.Tensor):
            match = True  # In case of an empty tensor.

            # If we have ``cls.dtype``, we can be maximally precise.
            if cls.dtype and cls.dtype != inst.dtype:
                match = False

            # Handle ellipses.
            else:
                shape_match, _ = shapecheck(inst.shape, cls.shape)
                attr_match, _ = attrcheck(inst, cls.kwattrs)
                match = shape_match and attr_match

        return match


class _Tensor(metaclass=_TensorMeta):
    """ This class exists to keep the Tensor class as clean as possible. """

    NAME: str = "Tensor"
    DIM_TYPES: List[type] = TORCH_DIM_TYPES
    _TORCH_DTYPE_MAP: Dict[type, torch.dtype] = TORCH_DTYPE_MAP

    dtype: Optional[torch.dtype] = None
    shape: Optional[Tuple] = None
    kwattrs: Optional[Dict[str, Any]] = None

    def __new__(cls, *args: Tuple[Any], **kwargs: Dict[str, Any]) -> Any:
        raise TypeError("Cannot instantiate abstract class 'Tensor'.")

    @classmethod
    def get_dtype(cls, item: Any) -> Tuple[Optional[torch.dtype], ...]:
        """ Computes dtype. """
        dtype = None

        # Case where ``item`` is a dtype (``Tensor[torch.float64]``).
        if isinstance(item, torch.dtype):
            dtype = item

        elif isinstance(item, np.dtype):
            np_dtype_err = f"Invalid type argument '{item}'. "
            np_dtype_err += f"Numpy dtypes not supported for Tensor class. "
            np_dtype_err += f"Type arguments must be torch dtypes or in "
            np_dtype_err += f"'{list(cls._TORCH_DTYPE_MAP.keys())}'."
            raise TypeError(np_dtype_err)

        # Case where ``item`` is a python3 type (``Tensor[int]``).
        elif isinstance(item, type):
            generic_type = item
            if generic_type not in cls._TORCH_DTYPE_MAP:
                invalid_type_err = f"Invalid type argument '{generic_type}'. "
                invalid_type_err += "Type arguments must be in "
                invalid_type_err += f"'{list(cls._TORCH_DTYPE_MAP.keys())}'."
                raise TypeError(invalid_type_err)
            dtype = cls._TORCH_DTYPE_MAP[generic_type]

        return dtype, None

    @classmethod
    def _after_subscription(
        cls, item: Union[type, Optional[Union[int, EllipsisType]]]  # type: ignore
    ) -> None:
        """ Set class attributes based on the passed dtype/dim data. """
        dtype, shape, kwattrs, _ = parse_subscript(cls, item, torch.dtype)
        cls.dtype = dtype
        cls.shape = shape
        cls.kwattrs = kwattrs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" This module contains a general subscript parser for subscriptable types. """
from typing import Optional, Tuple, Union, Dict, Any

from sympy.core.expr import Expr
from sympy.core.symbol import Symbol

from asta.utils import is_subtuple
from asta.scalar import Scalar
from asta.classes import SubscriptableMeta
from asta.constants import EllipsisType

# pylint: disable=unidiomatic-typecheck


def verify_dimensions(cls: SubscriptableMeta, item: Any) -> None:
    """ Raises a TypeError if the type of a dimension is invalid. """
    if not isinstance(item, tuple):
        item = (item,)

    for dim in item:

        # Treat sympy expressions specially.
        if isinstance(dim, (Expr, Symbol)):
            pass
        elif type(dim) not in cls.DIM_TYPES:
            err = f"Invalid dimension '{dim}' of type '{type(dim)}'. "
            err += f"Valid dimension types: {cls.DIM_TYPES}"
            raise TypeError(err)


def parse_subscript(
    cls: SubscriptableMeta,
    item: Union[type, Optional[Union[int, EllipsisType]]],  # type: ignore
    dtype_metaclass: type,
) -> Tuple[Optional[type], Optional[Tuple], Optional[Dict[str, Any]], str]:
    """ Set class attributes based on the passed dtype/dim data. """
    kind: str = ""
    if hasattr(cls, "kind"):
        kind = cls.kind
    dtype: Optional[type] = cls.dtype
    shape: Optional[Tuple] = cls.shape
    kwattrs: Optional[Dict[str, Any]] = cls.kwattrs

    if isinstance(item, (type, dtype_metaclass)) and item != Scalar:
        dtype, kind = cls.get_dtype(item)
        shape = None

    # Adding support for kwargs.
    elif isinstance(item, dict):
        kwattrs = item
        shape = None

    # Case where dtype is Any and shape is scalar.
    elif item in (Scalar, ()):
        shape = ()

    # Case where dtype is not passed in, and there's one input.
    # i.e. ``Array[1]`` or ``Array[...]``.
    elif not isinstance(item, tuple):

        verify_dimensions(cls, item)
        shape = (item,)

    # Case where ``item`` is a nonempty tuple.
    elif item:

        # Case where generic type is specified.
        if isinstance(item[0], (type, dtype_metaclass)) and item[0] != Scalar:
            dtype, kind = cls.get_dtype(item[0])
            item = item[1:]

        # Adding support for keyword attributes.
        if item and isinstance(item[-1], dict):
            kwattrs = item[-1]
            item = item[:-1]

        if item:
            verify_dimensions(cls, item)
            shape = SubscriptableMeta.get_shape(item)

    else:
        empty_err = "Argument to '{cls.NAME}[]' cannot be empty tuple. "
        empty_err += "Use '{cls.NAME}[None]' to indicate a scalar."
        raise TypeError(empty_err)

    if isinstance(shape, tuple) and is_subtuple((..., ...), shape, set())[0]:
        raise TypeError("Invalid shape: repeated '...'")

    return dtype, shape, kwattrs, kind



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" A module for programmatically storing dimension sizes for annotations. """
import sys
from typing import Dict, Optional, Any, Union
from sympy import symbols
from sympy.core.symbol import Symbol

# pylint: disable=redefined-outer-name, too-few-public-methods, no-self-use

# Dummy map to trick pylint.
symbol_map: Dict[Symbol, Optional[int]] = {}


def __getattr__(name: str) -> Any:
    """ This exists solely to trick pylint. """
    raise NotImplementedError


class Dimensions:
    """ An instance of this object acts as a proxy for this module. """

    def __init__(self) -> None:
        # Set any attributes here - before initialisation (they remain normal attrs).
        self.symbol_map: Dict[Symbol, Optional[int]] = {}

        # After initialization, setting attributes is the same as setting an item.
        self.__initialized = True

    def __getattr__(self, name: str) -> Union[Symbol, int]:
        symbol = symbols(name)
        if symbol not in self.symbol_map:
            self.symbol_map[symbol] = None
            return symbol

        value: Optional[int] = self.symbol_map[symbol]
        if value is None:
            return symbol
        assert isinstance(value, int)
        return value

    def __setattr__(self, name: str, value: Any) -> None:
        """ Maps attributes to values. Only if we are initialised. """
        # This test allows attributes to be set in the ``__init__()`` method.
        if "_Dimensions__initialized" not in self.__dict__:
            super().__setattr__(name, value)
        else:
            if not isinstance(value, int):
                raise TypeError("Value of a dim must be an integer.")
            symbol = symbols(name)
            self.symbol_map[symbol] = value


sys.modules[__name__] = Dimensions()  # type: ignore[assignment]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Constants and helper classes for asta types. """
import datetime
from typing import Dict, List, Union, Any

import numpy as np
from sympy.core.expr import Expr
from sympy.core.symbol import Symbol

from asta.placeholder import Placeholder

_TORCH_IMPORTED = False
try:
    import torch

    _TORCH_IMPORTED = True
except ImportError:
    pass
_TENSORFLOW_IMPORTED = False
try:
    import tensorflow as tf

    _TENSORFLOW_IMPORTED = True
except ImportError:
    pass


# pylint: disable=invalid-name, too-few-public-methods


# Classes and metaclasses.
class NonInstanceMeta(type):
    """ Metaclass for ``NonInstanceType``. """

    def __instancecheck__(cls, inst: Any) -> bool:
        """ No object is an instance of this type. """
        return False


class NonInstanceType(metaclass=NonInstanceMeta):
    """ No object is an instance of this class. """


class TorchModule:
    """ A dummy torch module for when torch is not installed. """

    def __init__(self) -> None:
        self.Tensor = NonInstanceType
        self.Size = NonInstanceType
        self.dtype = NonInstanceType
        self.int32 = NonInstanceType
        self.float32 = NonInstanceType
        self.bool = NonInstanceType
        self.uint8 = NonInstanceType


class DTypes:
    """ A dummy dtypes attribute object for when tf is not installed. """

    def __init__(self) -> None:
        self.DType = NonInstanceType


class TFModule:
    """ A dummy tensorflow module for when tf is not installed. """

    # pylint: disable=too-many-instance-attributes
    def __init__(self) -> None:
        self.Tensor = NonInstanceType
        self.TensorShape = NonInstanceType
        self.dtypes = DTypes()
        self.dtypes.DType = NonInstanceType
        self.bfloat16 = NonInstanceType
        self.bool = NonInstanceType
        self.complex = NonInstanceType
        self.complex128 = NonInstanceType
        self.complex64 = NonInstanceType
        self.double = NonInstanceType
        self.float16 = NonInstanceType
        self.float32 = NonInstanceType
        self.float64 = NonInstanceType
        self.half = NonInstanceType
        self.int16 = NonInstanceType
        self.int32 = NonInstanceType
        self.int64 = NonInstanceType
        self.int8 = NonInstanceType
        self.qint16 = NonInstanceType
        self.qint32 = NonInstanceType
        self.qint8 = NonInstanceType
        self.quint16 = NonInstanceType
        self.quint8 = NonInstanceType
        self.resource = NonInstanceType
        self.string = NonInstanceType
        self.uint16 = NonInstanceType
        self.uint32 = NonInstanceType
        self.uint64 = NonInstanceType
        self.uint8 = NonInstanceType
        self.variant = NonInstanceType


class ScalarMeta(type):
    """ A meta class for the ``Scalar`` class. """

    _GENERIC_TYPES: List[type]
    _ARRAY_TYPES: List[type]

    def __instancecheck__(cls, inst: Any) -> bool:
        """ Support expected behavior for ``isinstance(<number-like>, Scalar)``. """
        for arr_type in cls._ARRAY_TYPES:
            if isinstance(inst, arr_type):
                assert hasattr(inst, "shape")
                if inst.shape == tuple():  # type: ignore[attr-defined]
                    return True
                return False
        for generic_type in cls._GENERIC_TYPES:
            if isinstance(inst, generic_type):
                return True
        return False


class Printable:
    """ Class for printing object representations in nested objects. """

    def __init__(self, rep: str):
        self.rep = rep

    def __repr__(self) -> str:
        """ Just return ``rep``. """
        return self.rep


class Color:
    """ Terminal color string literals. """

    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


if not _TORCH_IMPORTED:
    torch = TorchModule()
    tf = TFModule()

# Types.
GenericArray = Union[np.ndarray, torch.Tensor, tf.Tensor]
ARRAY_TYPES: List[type] = [np.ndarray, torch.Tensor, tf.Tensor]
GENERIC_TYPES: List[type] = [
    bool,
    int,
    float,
    complex,
    bytes,
    str,
    datetime.datetime,
    datetime.timedelta,
]
NoneType = type(None)
EllipsisType = type(Ellipsis)

CORE_DIM_TYPES: List[type] = [
    int,
    ScalarMeta,
    EllipsisType,
    NoneType,  # type: ignore[misc]
    tuple,
    Placeholder,
    Expr,
    Symbol,
]
NUMPY_DIM_TYPES: List[type] = CORE_DIM_TYPES
TORCH_DIM_TYPES: List[type] = CORE_DIM_TYPES + [torch.Size]
TF_DIM_TYPES: List[type] = CORE_DIM_TYPES + [tf.TensorShape]
ALL_DIM_TYPES: List[type] = CORE_DIM_TYPES + [torch.Size, tf.TensorShape]
NP_UNSIZED_TYPE_KINDS: Dict[type, str] = {bytes: "S", str: "U", object: "O"}
TORCH_DTYPE_MAP: Dict[type, torch.dtype] = {
    int: torch.int32,
    float: torch.float32,
    bool: torch.bool,
    bytes: torch.uint8,
}
TF_DTYPE_MAP: Dict[type, tf.dtypes.DType] = {
    int: tf.int32,
    float: tf.float32,
    complex: tf.complex128,
    bool: tf.bool,
    bytes: tf.string,
    str: tf.string,
}
TF_DTYPES = [
    tf.bfloat16,
    tf.bool,
    tf.complex,
    tf.complex128,
    tf.complex64,
    tf.double,
    tf.float16,
    tf.float32,
    tf.float64,
    tf.half,
    tf.int16,
    tf.int32,
    tf.int64,
    tf.int8,
    tf.qint16,
    tf.qint32,
    tf.qint8,
    tf.quint16,
    tf.quint8,
    tf.resource,
    tf.string,
    tf.uint16,
    tf.uint32,
    tf.uint64,
    tf.uint8,
    tf.variant,
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Typechecking utilities. """
import random
import functools
from typing import List, Dict, Tuple, Union, Any, Set, Optional

from sympy import solvers, simplify
from sympy.core.expr import Expr
from sympy.core.symbol import Symbol
from sympy.core.numbers import Number
from asta.constants import (
    EllipsisType,
    NonInstanceType,
    GenericArray,
    torch,
    tf,
    _TORCH_IMPORTED,
    _TENSORFLOW_IMPORTED,
)

# pylint: disable=too-many-boolean-expressions, too-many-branches


def shapecheck(
    inst_shape: Tuple[int, ...],
    cls_shape: Optional[Tuple[Union[int, EllipsisType], ...]],  # type: ignore[valid-type]
) -> Tuple[bool, Set[Expr]]:
    """ Check ``inst_shape`` is an instance of ``cls_shape``. """
    match = True
    equations: Set[Expr] = set()

    if cls_shape is None:
        return match, equations

    assert isinstance(inst_shape, tuple)

    # The portions of ``inst_shape`` which correspond to each ``cls_shape`` elem.
    shape_pieces: List[Tuple[int, ...]] = []

    # Case 1: No ellipses or wildcards.
    if Ellipsis not in cls_shape and -1 not in cls_shape:
        equal, equations = check_equal(cls_shape, inst_shape, equations)

        if not equal:
            match = False
        shape_pieces = [(elem,) for elem in inst_shape]

    # Case 2: Only an ellipsis, and instance shape is empty.
    elif cls_shape == (Ellipsis,) and inst_shape == ():
        match = True
        shape_pieces = [tuple()]

    # Case 3: Nonempty, not an ellipsis, and instance shape is empty.
    elif inst_shape == () != cls_shape:
        match = False

    # Case 4: Ellipsis, and instance shape has a zero (empty tensor).
    elif cls_shape == (Ellipsis,) and 0 in inst_shape:
        match = False

    # Case 5: Everything else.
    else:
        if is_subtuple((Ellipsis, Ellipsis), cls_shape, equations)[0]:
            raise TypeError("Invalid shape: repeated '...'")

        # Determine if/where '...' bookends ``cls_shape``.
        left_bookend = False
        right_bookend = False
        ellipsis_positions: List[int] = []
        for i, elem in enumerate(cls_shape):
            if elem == Ellipsis:
                ellipsis_positions.append(i)

                # e.g. ``Array[..., 1, 2, 3]``.
                if i == 0:
                    left_bookend = True

                # e.g. ``Array[1, 2, 3, ...]``.
                if i == len(cls_shape) - 1:
                    right_bookend = True

        # Analogous to ``str.split(<elem>)``, we split the shape on '...'.
        frags: List[Tuple[int, ...]] = split(cls_shape, Ellipsis)

        # Index in ``cls_shape`` as we construct ``shape_pieces``.
        cls_idx = 0

        # ``ishape`` is the remaining portion of the instance shape.
        ishape = inst_shape
        for i, frag in enumerate(frags):

            # Look for ``frag`` in ``ishape``, and find starting index.
            is_sub, index, equations = is_subtuple(frag, ishape, equations)

            # Must have ``frag`` contained in ``ishape``.
            if not is_sub:
                match = False
                break

            # First fragment must start at 0 if '...' is not the first
            # element of ``cls_shape``.
            if i == 0 and not left_bookend and index != 0:
                match = False
                break

            # Last fragement must end at (exclusive) ``len(ishape)`` if
            # '...' is not the last element of ``cls_shape``.
            if (
                i == len(frags) - 1
                and not right_bookend
                and index + len(frag) != len(ishape)
            ):
                match = False
                break

            # The subseq of ``ishape`` being substituted for ``...``.
            substituted = ishape[:index]

            # Can't match zero-size dims with ``...``.
            if 0 in substituted:
                match = False
                break

            cls_elem = cls_shape[cls_idx]
            if cls_elem == Ellipsis:
                shape_pieces.append(substituted)
                cls_idx += 1

            ifrag = ishape[index : index + len(frag)]
            for elem in ifrag:
                shape_pieces.append((elem,))
                cls_idx += 1

            new_start = index + len(frag)
            ishape = ishape[new_start:]

        # If ``cls_shape`` ends with ``...``, ensure substituted tuple has no 0s.
        if right_bookend and 0 in ishape:
            match = False

        if right_bookend:
            substituted = ishape
            cls_elem = cls_shape[cls_idx]
            if cls_elem == Ellipsis:
                shape_pieces.append(substituted)
                cls_idx += 1

        if match:
            reconstructed = functools.reduce(lambda a, b: a + b, shape_pieces)
            assert reconstructed == inst_shape
            assert len(shape_pieces) == len(cls_shape) == cls_idx

    # Solve our system of equations if it is nonempty.
    if match:
        match, _, _ = astasolver(equations)

    return match, equations


def attrcheck(
    inst: GenericArray, kwattrs: Optional[Dict[str, Any]]
) -> Tuple[bool, Set[Expr]]:
    """ Check if ``inst`` has attributes matching ``kwattrs``. """
    match = True
    equations: Set[Expr] = set()
    kwattrs = {} if kwattrs is None else kwattrs

    assert isinstance(kwattrs, dict)

    for key, value in kwattrs.items():
        attr = getattr(inst, key, NonInstanceType)
        if attr is NonInstanceType:
            match = False
            break
        if isinstance(value, Expr):
            equations.add(value - attr)
        elif attr != value:
            match = False
            break

    return match, equations


def astasolver(
    equations: Set[Expr],
) -> Tuple[bool, Set[Symbol], List[Dict[Symbol, int]]]:
    """ Solve equations for positive size free symbols. """
    # Treat case where there are no symbols, and the equation is ``0 = 0``.
    if 0 in equations:
        equations.remove(0)

    # If there are no nontrivial equations, return True.
    if not equations:
        return True, set(), []

    symbols: Set[Symbol] = set()
    for equation in equations:
        symbols = symbols.union(equation.free_symbols)
    solutions: List[Dict[Symbol, int]] = solvers.solve(equations, symbols, dict=True)

    # Remove symbolic solutions, e.g. ``X = Y``.
    pruned_solutions: List[Dict[Symbol, Expr]] = []
    for solution in solutions:
        all_numbers = True
        for _key, val in solution.items():
            if not isinstance(val, Number):
                all_numbers = False
                break
        if all_numbers:
            pruned_solutions.append(solution)

    # If we don't get at least one solution, it's not a match.
    return len(pruned_solutions) >= 1, symbols, pruned_solutions


def is_subtuple(
    sub: Tuple[Union[int, EllipsisType], ...],  # type: ignore[valid-type]
    tup: Tuple[Union[int, EllipsisType], ...],  # type: ignore[valid-type]
    equations: Set[Expr],
) -> Tuple[bool, int, Set[Expr]]:
    """ Check for tuple inclusion, return index of first one. """
    assert isinstance(sub, tuple)
    assert isinstance(tup, tuple)
    for i in range(len(tup) - len(sub) + 1):
        equal, equations = check_equal(sub, tup[i : i + len(sub)], equations)
        if equal:
            return True, i, equations
    return False, -1, equations


def split(
    shape: Tuple[Union[int, EllipsisType], ...],  # type: ignore[valid-type]
    elem: Union[int, EllipsisType],  # type: ignore[valid-type]
) -> List[Tuple[int, ...]]:
    """ Split on an element. """
    shape_list = list(shape)
    tokens: List[str] = []
    for num in shape_list:
        if num == elem:
            tokens.append("@")
        else:
            assert isinstance(num, int)
            tokens.append(str(num))
    seq: str = ",".join(tokens)
    comma_fragments: List[str] = seq.split("@")

    result: List[Tuple[int, ...]] = []
    for frag in comma_fragments:
        if frag:
            frag_tokens = [token for token in frag.split(",") if token]
            frag_nums = [int(token) for token in frag_tokens]
            frag_tuple = tuple(frag_nums)
            result.append(frag_tuple)

    return result


def check_equal(
    shape_1: Tuple[Union[int, EllipsisType], ...],  # type: ignore[valid-type]
    shape_2: Tuple[Union[int, EllipsisType], ...],  # type: ignore[valid-type]
    equations: Set[Expr],
) -> Tuple[bool, Set[Expr]]:
    """
    Determines if two shape tuples are equal, allowing wildcards (``-1``),
    which can take the place of an positive integer, and equations, which can take
    the place of a fixed positive integer. They are set to the first value
    matched and fixed within a single isinstance check and within a single
    function typecheck operation. We also allow Ellipses, but these are treated
    as atomic elements.
    """
    if len(shape_1) != len(shape_2):
        return False, equations
    for x, y in zip(shape_1, shape_2):

        # Case 1: Both are integer literals.
        if (
            isinstance(x, int)
            and isinstance(y, int)
            and ((x == -1 and y != 0) or (y == -1 and x != 0))
        ):
            continue

        # Case 2: ``x`` is an expression.
        if isinstance(x, Expr) and isinstance(y, int):
            equations.add(x - y)
            continue

        # Case 3: ``y`` is an expression.
        if isinstance(y, Expr) and isinstance(x, int):
            equations.add(y - x)
            continue

        if isinstance(x, Expr) and isinstance(y, Expr):

            if simplify(x - y) != 0:
                return False, equations
            continue

        if x != y:
            return False, equations
    return True, equations


def rand_split_shape(shape: Any) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
    """ Splits a shape and removes a nonempty continguous portion. """
    if _TORCH_IMPORTED and isinstance(shape, torch.Size):
        shape = tuple(shape)
    if _TENSORFLOW_IMPORTED and isinstance(shape, tf.TensorShape):
        shape = tuple(shape)
    start = random.randrange(len(shape))
    end = random.randint(start + 1, len(shape))
    left = shape[:start]
    right = shape[end:]
    return left, right


def shape_repr(shape: Tuple[int, ...]) -> str:
    """ Get stripped representation of a shape. """
    return repr(shape).replace("Ellipsis", "...")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Support for typing Numpy ndarrays. """
import numpy as np
from asta._array import _Array


class Array(_Array, np.ndarray):
    """
    A subclass of ``np.ndarray`` for use in type annotations.

    Example of an array with an undefined generic type and shape:
        ``Array``

    Example of an array with a defined generic type:
        ``Array[int]``

    Example of an array with a defined numpy dtype:
        ``Array[np.int32]``

    Example of an array with a defined numpy dtype and shape:
        ``Array[np.int32, 3]``
        ``Array[np.int64, 1, 2, 3]``

    Example of an array with wildcard dimension.
        ``Array[-1]``
        ``Array[int, 1, 2, -1, 3]``

    Example of an array with ellipses.
        ``Array[...]``
        ``Array[int, ...]``
        ``Array[int, 1, 2, ...]``
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Support for typing PyTorch tensors. """
import torch
from asta._tensor import _Tensor


class Tensor(_Tensor, torch.Tensor):
    """
    A subclass of ``torch.Tensor`` for use in type annotations.

    Example of an tensor with an undefined generic type and shape:
        ``Tensor``

    Example of an tensor with a defined generic type:
        ``Tensor[int]``

    Example of an tensor with a defined torch dtype:
        ``Tensor[torch.int32]``

    Example of an tensor with a defined torch dtype and shape:
        ``Tensor[torch.int32, 3]``
        ``Tensor[torch.int64, 1, 2, 3]``

    Example of an tensor with wildcard dimension.
        ``Tensor[-1]``
        ``Tensor[int, 1, 2, -1, 3]``

    Example of an tensor with ellipses.
        ``Tensor[...]``
        ``Tensor[int, ...]``
        ``Tensor[int, 1, 2, ...]``
    """

    def __ipow__(self, other: torch.Tensor) -> torch.Tensor:
        """ In-place augmented arithmetic method for exponentiation (``**=``). """
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" A placeholder class for lazy-set shapes in asta annotations. """
from typing import Optional, List, Union, Tuple, Any
from sympy.core.expr import Expr
from sympy.core.symbol import Symbol

# pylint: disable=no-self-use


class Placeholder:
    """ Placeholder for annotation dimensions. """

    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name
        self.unpacked = False
        self.composite = False
        self.contents: List[Union[Placeholder, Tuple[int, ...]]] = []

    def __repr__(self) -> str:
        """ String representation of placeholder. """
        content = f"[{self.contents}]" if self.contents else ""
        name = f"{self.name}" if self.name is not None else ""
        return f"#{name}{content}"

    def __iter__(self) -> object:
        """ Make sure instances can be unpacked. """
        self.unpacked = True
        yield self

    def __next__(self) -> None:
        """ Instances are empty iterators. """
        raise StopIteration

    def __add__(self, summand: Any) -> Any:
        """ Return the concatenation. """
        left_contents = self.contents if self.contents else [self]
        if isinstance(summand, tuple):
            for elem in summand:
                if not isinstance(elem, (int, Expr, Symbol)):
                    raise TypeError(
                        "Shape elements must be integers, symbols, or expressions."
                    )
            sum_contents = left_contents + [summand]
        elif isinstance(summand, Placeholder):
            right_contents = summand.contents if summand.contents else [summand]
            sum_contents = left_contents + right_contents
        else:
            operand_err = "Unsupported operand type(s) for +: "
            operand_err += f"'Placeholder' and '{type(summand)}'"
            raise TypeError(operand_err)

        # Construct the sum placeholder.
        result = Placeholder()
        result.contents = sum_contents
        result.composite = True
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Dummy classes for use when ``torch`` or ``tensorflow`` are not installed. """
from typing import Any
from abc import abstractmethod
from asta.classes import SubscriptableMeta, GenericMeta

# pylint: disable=import-outside-toplevel, unused-import, too-few-public-methods


class UnusableMeta(SubscriptableMeta):
    """ A meta class for the dummy ``Tensor`` and ``TFTensor`` classes. """

    @classmethod
    @abstractmethod
    def _after_subscription(cls, item: Any) -> None:
        """ Method signature for subscript argument processing. """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _import(cls) -> None:
        """ Attempts to import the dependency so a meaningful ImportError is shown. """
        raise NotImplementedError

    def __getitem__(cls, item: Any) -> GenericMeta:
        """ Should always raise an ImportError. """
        cls._import()
        return SubscriptableMeta.__getitem__(cls, item)

    def __eq__(cls, other: Any) -> bool:
        """ Should always raise an ImportError. """
        cls._import()
        return super().__eq__(other)

    def __repr__(cls) -> str:
        """ Should always raise an ImportError. """
        cls._import()
        return super().__repr__()

    def __instancecheck__(cls, inst: Any) -> bool:
        """ Should always raise an ImportError. """
        cls._import()
        return False


class Tensor(metaclass=UnusableMeta):
    """ A dummy class for use when ``torch`` is not installed. """

    @classmethod
    def _import(cls) -> None:
        """ Attempts to import the dependency so a meaningful ImportError is shown. """
        import torch


class TFTensor(metaclass=UnusableMeta):
    """ A dummy class for use when ``tensorflow`` is not installed. """

    @classmethod
    def _import(cls) -> None:
        """ Attempts to import the dependency so a meaningful ImportError is shown. """
        import tensorflow



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Supermetaclasses for ``_Array``-like asta types. """
import types
from abc import abstractmethod
from typing import TypeVar, Generic, Any, Optional, Tuple, List, Dict

import numpy as np
from oxentiel import Oxentiel

from asta.utils import shape_repr
from asta.config import get_ox
from asta.scalar import Scalar
from asta.constants import Printable

# pylint: disable=too-few-public-methods

T = TypeVar("T")


class GenericMeta(type, Generic[T]):
    """ Abstract base metaclass for subscriptable types. """

    kind: str
    shape: tuple
    dtype: Any
    kwattrs: Dict[str, Any]

    @classmethod
    @abstractmethod
    def _after_subscription(cls, item: Any) -> None:
        """ Method signature for subscript argument processing. """
        raise NotImplementedError


class SubscriptableMeta(GenericMeta):
    """
    This metaclass will allow a type to become subscriptable.

    >>> class SomeType(metaclass=GenericMeta):
    ...     pass
    >>> SomeTypeSub = SomeType['some args']
    >>> SomeTypeSub.__args__
    'some args'
    >>> SomeTypeSub.__origin__.__name__
    'SomeType'
    """

    DIM_TYPES: List[type]
    NAME: str
    OX: Oxentiel

    __args__: Any
    __origin__: Any

    def __init_subclass__(cls) -> None:
        cls._hash = 0
        cls.__args__ = None
        cls.__origin__ = None

    def __init__(cls, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]):
        """ Initializes the configuration object if it doesn't already exist. """
        super().__init__(name, bases, attrs)
        ox = get_ox()
        cls.OX = ox

    @classmethod
    @abstractmethod
    def _after_subscription(cls, item: Any) -> None:
        """ Method signature for subscript argument processing. """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_dtype(item: Any) -> Tuple[Optional[np.dtype], str]:
        """ Computes dtype. """
        raise NotImplementedError

    def __getitem__(cls, item: Any) -> GenericMeta:
        body = {
            **cls.__dict__,
            "__args__": item,
            "__origin__": cls,
        }
        bases = cls, *cls.__bases__
        result: GenericMeta = type(cls.__name__, bases, body)  # type: ignore

        if hasattr(result, "_after_subscription"):

            # Verify it is not a staticmethod.
            if isinstance(result._after_subscription, types.FunctionType):
                name = "_after_subscription"
                static_err = f"The '{name}' method should not be static."
                raise TypeError(static_err)

            result._after_subscription(item)
        return result

    def __eq__(cls, other: Any) -> bool:
        cls_args = getattr(cls, "__args__", None)
        cls_origin = getattr(cls, "__origin__", None)
        other_args = getattr(other, "__args__", None)
        other_origin = getattr(other, "__origin__", None)
        eq: bool = cls_args == other_args and cls_origin == other_origin
        return eq

    def __hash__(cls) -> int:
        if not getattr(cls, "_hash", None):
            cls._hash = hash("{}{}".format(cls.__origin__, cls.__args__))
        return cls._hash

    def __repr__(cls) -> str:
        """ String representation of class. """
        assert hasattr(cls, "shape")
        assert hasattr(cls, "dtype")
        assert hasattr(cls, "kwattrs")
        subscript: List[Any] = []
        if cls.dtype is not None:
            if cls.NAME == "Array":
                printable_dtype = Printable(f"np.{cls.dtype.name}")
                subscript.append(printable_dtype)
            else:
                subscript.append(cls.dtype)
        if cls.shape is not None:
            shape = tuple(cls.shape)
            printable_shape = Printable(f"shape={shape_repr(shape)}")
            subscript.append(printable_shape)
        if cls.kwattrs is not None:
            printable_kwattrs = Printable(f"attrs={cls.kwattrs}")
            subscript.append(printable_kwattrs)

        rep = f"<asta.{cls.NAME}{subscript}>"

        return rep

    @staticmethod
    def get_shape(item: Tuple) -> Optional[Tuple]:
        """ Compute shape from a shape tuple argument. """
        shape: Optional[Tuple] = None

        if item != tuple() and item is not None:
            if Scalar not in item and () not in item:
                shape = item
            elif item in [(Scalar,), ((),)]:
                shape = ()
            else:
                none_err = "Too many 'None' arguments. "
                none_err += "Use 'Array[None]' for scalar arrays."
                raise TypeError(none_err)

        # ``((1,2,3),)`` -> ``(1,2,3)``.
        if isinstance(shape, tuple) and len(shape) == 1 and isinstance(shape[0], tuple):
            shape = shape[0]

        return shape



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Support for typing TensorFlow tensors. """
import tensorflow as tf
from asta._tftensor import _TFTensor

# pylint: disable=too-few-public-methods


class TFTensor(_TFTensor, tf.Tensor):
    """
    A subclass of ``tf.Tensor`` for use in type annotations.

    Example of an tensor with an undefined generic type and shape:
        ``Tensor``

    Example of an tensor with a defined generic type:
        ``Tensor[int]``

    Example of an tensor with a defined tf dtype:
        ``Tensor[tf.int32]``

    Example of an tensor with a defined tf dtype and shape:
        ``Tensor[tf.int32, 3]``
        ``Tensor[tf.int64, 1, 2, 3]``

    Example of an tensor with wildcard dimension.
        ``Tensor[-1]``
        ``Tensor[int, 1, 2, -1, 3]``

    Example of an tensor with ellipses.
        ``Tensor[...]``
        ``Tensor[int, ...]``
        ``Tensor[int, 1, 2, ...]``
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Scalar type class. """
from typing import List
from asta.constants import GENERIC_TYPES, ARRAY_TYPES, ScalarMeta

# pylint: disable=too-few-public-methods


class Scalar(metaclass=ScalarMeta):
    """ A generic scalar type class. """

    _GENERIC_TYPES: List[type] = GENERIC_TYPES
    _ARRAY_TYPES: List[type] = ARRAY_TYPES



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Defines the ``@typechecked`` decorator. """
import os
import inspect
from typing import Any, Tuple, Dict, Set

from sympy.core.expr import Expr

from oxentiel import Oxentiel

from asta.utils import astasolver
from asta.config import get_ox
from asta.origins import check_annotation
from asta.display import (
    fail_system,
    get_header,
    handle_pass,
)


def validate_annotations(  # type: ignore[no-untyped-def]
    decorated, annotations: Dict[str, Any], args: Tuple[Any], kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Make sure there is an annotation for each parameter, return arguments in
    args other than class references (self, cls, mcs), and keyword arguments in
    kwargs with defaults filled in.
    """
    num_annots = len(annotations)
    num_non_return_annots = num_annots
    if "return" in annotations:
        num_non_return_annots -= 1

    # Get the parameter list from the function signature.
    sig = inspect.signature(decorated)
    paramlist = list(sig.parameters)

    # Get a list of kwargs with defaults filled-in.
    defaults: Dict[str, Any] = {}
    for k, v in sig.parameters.items():
        if v.default is not inspect.Parameter.empty:
            defaults[k] = v.default

    checkable_args: Dict[str, Any] = defaults.copy()
    checkable_args.update(kwargs)

    # Remove unannotated instance/class/metaclass reference.
    pure_args = args
    refs = ("self", "cls", "mcs")

    if len(sig.parameters) == num_non_return_annots + 1 and paramlist[0] in refs:
        pure_args = pure_args[1:]  # type: ignore[assignment]
    for i, arg in enumerate(pure_args):
        name = list(annotations.keys())[i]
        checkable_args[name] = arg

    # Check for mismatch between lengths of arguments/annotations.
    if num_non_return_annots != len(checkable_args):
        num_annot_err = f"Mismatch between number of annotated "
        num_annot_err += f"non-(self / cls / mcs) parameters "
        num_annot_err += f"'({num_non_return_annots})' and number of arguments "
        num_annot_err += f"'({len(checkable_args)})'. "
        num_annot_err += f"There may be a type annotation missing."
        raise TypeError(num_annot_err)

    return checkable_args


def typechecked(decorated):  # type: ignore[no-untyped-def]
    """
    Typecheck a function annotated with ``asta`` type objects. This decorator
    will only check the shape and datatype of parameters annotated with
    ``asta`` type variables.  Mypy should be used for everything else. Note
    that the argument-parameter assignment problem is trivial because the only
    arguments which can be out of order are keyword arguments.

    Parameters
    ----------
    decorated : ``Callable[[Any], Any]``.
        The function to be typechecked.

    Returns
    -------
    _wrapper : ``Callable[[Any], Any]``.
        The decorated version of ``decorated``.
    """
    ox: Oxentiel = get_ox()
    if "ASTA_TYPECHECK" in os.environ:
        ox.on = ox.on and os.environ["ASTA_TYPECHECK"] == "1"
    if not ox.on:
        return decorated

    # Treat classes.
    if inspect.isclass(decorated):

        # Grab the module name.
        prefix = decorated.__qualname__ + "."

        # Iterate over attributes.
        for key, attr in decorated.__dict__.items():

            # If it's decoratable.
            if (
                inspect.isfunction(attr)
                or inspect.ismethod(attr)
                or inspect.isclass(attr)
            ):
                # If the name prefix matches and it has annotations.
                if attr.__qualname__.startswith(prefix) and getattr(
                    attr, "__annotations__", None
                ):

                    # Decorate the method/function/class.
                    setattr(decorated, key, typechecked(attr))

            # Only for class and staticmethods; instance methods are caught above.
            elif isinstance(attr, (classmethod, staticmethod)):

                # If the underlying function has annotations.
                if getattr(attr.__func__, "__annotations__", None):
                    wrapped = typechecked(attr.__func__)

                    # Re-wrap with ``classmethod`` or ``staticmethod`` and put back.
                    setattr(decorated, key, type(attr)(wrapped))

        return decorated

    def _wrapper(*args: Tuple[Any], **kwargs: Dict[str, Any]) -> Any:
        """ Decorated/typechecked function. """

        # Print header for ``decorated``.
        ox.decorated = decorated
        header: str = get_header(decorated)
        handle_pass(header, ox)

        equations: Set[Expr] = set()
        annotations: Dict[str, Any] = decorated.__annotations__
        checkable_args: Dict[str, Any] = validate_annotations(
            decorated,
            annotations,
            args,  # type: ignore
            kwargs,
        )

        # Check arguments.
        for name, arg in checkable_args.items():
            annotation = annotations[name]
            equations = check_annotation(name, arg, annotation, equations, ox)
            del annotation

        # Call the decorated function.
        ret = decorated(*args, **kwargs)

        # Check return.
        ox.decorated = decorated
        if "return" in annotations:
            annotation = annotations["return"]
            equations = check_annotation("return", ret, annotation, equations, ox)
            del annotation

        # Solve our system of equations if it is nonempty.
        solvable, symbols, solutions = astasolver(equations)
        if not solvable:
            fail_system(equations, symbols, solutions, ox)

        return ret

    _wrapper.__module__ = decorated.__module__
    _wrapper.__name__ = decorated.__name__

    return _wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Simple check list from AllenNLP repo:
https://github.com/allenai/allennlp/blob/master/setup.py

To create the package for pypi.

1.  Change the version in __init__.py, setup.py as well as docs/source/conf.py.

2.  Commit these changes with the message: "Release: VERSION"

3.  Add a tag in git to mark the release:
        git tag VERSION -m 'Adds tag VERSION for pypi'

    Push the tag to git:
        git push --tags origin master

4.  Build both the sources and the wheel.

    Do not change anything in setup.py between creating the wheel and the
    source distribution (obviously).

    For the wheel, run: "python setup.py bdist_wheel" in the top level
    directory.  (this will build a wheel for the python version you use to
    build it).

    For the sources, run: "python setup.py sdist" You should now have a /dist
    directory with both .whl and .tar.gz source versions.

5.  Check that everything looks correct by uploading package to test server:

    twine upload dist/* -r pypitest

    (pypi suggest using twine as other methods upload files via plaintext.)

    You may have to specify the repository url,
    use the following command then:
        twine upload dist/* -r pypitest\
        --repository-url=https://test.pypi.org/legacy/

    Check that you can install it in a virtualenv by running:
        pip install -i https://testpypi.python.org/pypi transformers

6.  Upload the final version to actual pypi:

    twine upload dist/* -r pypi

7.  Copy the release notes from RELEASE.md to the tag in github.

"""
import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname: str) -> str:
    """ Read and return README as str. """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="asta",
    version="0.0.4",
    author="Brendan Whitaker",
    description=("Shape annotations for homogeneous numpy arrays and pytorch/tensorflow tensors."),
    license="GPLv3",
    packages=["asta"],
    long_description=read("README"),
    long_description_content_type="text/plain",
    install_requires=["numpy", "sympy", "oxentiel"],
    package_data={"asta": ["asta/defaults/astarc"]},
    include_package_data=True,
    python_requires=">=3.7.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.7",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" A switchboard for importing asta modules with large dependencies. """
# pylint: disable=unused-import, reimported, invalid-name
from asta.unusable import Tensor, TFTensor
from asta.constants import _TORCH_IMPORTED, _TENSORFLOW_IMPORTED


if _TORCH_IMPORTED:
    from asta.tensor import Tensor
if _TENSORFLOW_IMPORTED:
    from asta.tftensor import TFTensor
